package com.orientechnologies.orient.core.storage.impl.local.paginated;

import static org.assertj.core.api.Assertions.assertThat;


import com.orientechnologies.common.directmemory.OByteBufferPool;
import com.orientechnologies.common.directmemory.OPointer;
import com.orientechnologies.orient.core.record.ORecordVersionHelper;
import com.orientechnologies.orient.core.storage.cache.OCacheEntry;
import com.orientechnologies.orient.core.storage.cache.OCacheEntryImpl;
import com.orientechnologies.orient.core.storage.cache.OCachePointer;
import com.orientechnologies.orient.core.storage.cluster.OClusterPage;
import com.orientechnologies.orient.core.storage.impl.local.paginated.atomicoperations.OCacheEntryChanges;
import com.orientechnologies.orient.core.storage.impl.local.paginated.wal.OWALChanges;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import org.junit.Assert;
import org.junit.Test;

/**
 * @author Andrey Lomakin (a.lomakin-at-orientdb.com)
 * @since 20.03.13
 */
public class ClusterPageTest {
  private static final int SYSTEM_OFFSET = 24;

  @Test
  public void testAddOneRecord() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      addOneRecord(localPage);
      addOneRecord(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void addOneRecord(OClusterPage localPage) {
    int freeSpace = localPage.getFreeSpace();
    Assert.assertEquals(localPage.getRecordsCount(), 0);

    int recordVersion = 1;

    int position =
        localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});
    Assert.assertEquals(localPage.getRecordsCount(), 1);
    Assert.assertEquals(localPage.getRecordSize(0), 11);
    Assert.assertEquals(position, 0);
    Assert.assertEquals(localPage.getFreeSpace(),
        freeSpace - (27 + ORecordVersionHelper.SERIALIZED_SIZE));
    Assert.assertFalse(localPage.isDeleted(0));
    Assert.assertEquals(localPage.getRecordVersion(0), recordVersion);

    //    Assert.assertEquals(localPage.getRecordBinaryValue(0, 0, 11), new byte[] { 1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1 });

    assertThat(localPage.getRecordBinaryValue(0, 0, 11)).isEqualTo(new byte[] { 1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1 });
  }

  @Test
  public void testAddThreeRecords() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer,
        false);
    directCacheEntry.acquireExclusiveLock();

    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      addThreeRecords(localPage);
      addThreeRecords(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void addThreeRecords(OClusterPage localPage) {
    int freeSpace = localPage.getFreeSpace();

    Assert.assertEquals(localPage.getRecordsCount(), 0);

    int recordVersion = 0;
    recordVersion++;

    int positionOne =
        localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});
    int positionTwo =
        localPage.appendRecord(recordVersion, new byte[] {2, 2, 3, 4, 5, 6, 5, 4, 3, 2, 2});
    int positionThree =
        localPage.appendRecord(recordVersion, new byte[] {3, 2, 3, 4, 5, 6, 5, 4, 3, 2, 3});

    Assert.assertEquals(localPage.getRecordsCount(), 3);
    Assert.assertEquals(positionOne, 0);
    Assert.assertEquals(positionTwo, 1);
    Assert.assertEquals(positionThree, 2);

    Assert.assertEquals(localPage.getFreeSpace(), freeSpace - (3 * (27 + ORecordVersionHelper.SERIALIZED_SIZE)));
    Assert.assertFalse(localPage.isDeleted(0));
    Assert.assertFalse(localPage.isDeleted(1));
    Assert.assertFalse(localPage.isDeleted(2));

    //    Assert.assertEquals(localPage.getRecordBinaryValue(0, 0, 11), new byte[] { 1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1 });
    assertThat(localPage.getRecordBinaryValue(0, 0, 11)).isEqualTo(new byte[] { 1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1 });
    Assert.assertEquals(localPage.getRecordSize(0), 11);
    Assert.assertEquals(localPage.getRecordVersion(0), recordVersion);

    //    Assert.assertEquals(localPage.getRecordBinaryValue(1, 0, 11), new byte[] { 2, 2, 3, 4, 5, 6, 5, 4, 3, 2, 2 });
    assertThat(localPage.getRecordBinaryValue(1, 0, 11)).isEqualTo(new byte[] { 2, 2, 3, 4, 5, 6, 5, 4, 3, 2, 2 });
    Assert.assertEquals(localPage.getRecordSize(0), 11);
    Assert.assertEquals(localPage.getRecordVersion(1), recordVersion);

    //    Assert.assertEquals(localPage.getRecordBinaryValue(2, 0, 11), new byte[] { 3, 2, 3, 4, 5, 6, 5, 4, 3, 2, 3 });
    assertThat(localPage.getRecordBinaryValue(2, 0, 11)).isEqualTo(new byte[] { 3, 2, 3, 4, 5, 6, 5, 4, 3, 2, 3 });
    Assert.assertEquals(localPage.getRecordSize(0), 11);
    Assert.assertEquals(localPage.getRecordVersion(2), recordVersion);
  }

  @Test
  public void testAddFullPage() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer,
        false);
    directCacheEntry.acquireExclusiveLock();

    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      addFullPage(localPage);
      addFullPage(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void addFullPage(OClusterPage localPage) {
    int recordVersion = 0;
    recordVersion++;

    List<Integer> positions = new ArrayList<>();
    int lastPosition;
    byte counter = 0;
    int freeSpace = localPage.getFreeSpace();
    do {
      lastPosition = localPage.appendRecord(recordVersion, new byte[] {counter, counter, counter});
      if (lastPosition >= 0) {
        Assert.assertEquals(lastPosition, positions.size());
        positions.add(lastPosition);
        counter++;

        Assert.assertEquals(localPage.getFreeSpace(), freeSpace - (19 + ORecordVersionHelper.SERIALIZED_SIZE));
        freeSpace = localPage.getFreeSpace();
      }
    } while (lastPosition >= 0);

    Assert.assertEquals(localPage.getRecordsCount(), positions.size());

    counter = 0;
    for (int position : positions) {
      //      Assert.assertEquals(localPage.getRecordBinaryValue(position, 0, 3), new byte[] { counter, counter, counter });
      assertThat(localPage.getRecordBinaryValue(position, 0, 3)).isEqualTo(new byte[] { counter, counter, counter });
      Assert.assertEquals(localPage.getRecordSize(position), 3);
      Assert.assertEquals(localPage.getRecordVersion(position), recordVersion);
      counter++;
    }
  }

  @Test
  public void testDeleteAddLowerVersion() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer,
        false);
    directCacheEntry.acquireExclusiveLock();

    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      deleteAddLowerVersion(localPage);
      deleteAddLowerVersion(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void deleteAddLowerVersion(OClusterPage localPage) {
    int recordVersion = 0;
    recordVersion++;
    recordVersion++;

    int position =
        localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});

    Assert.assertTrue(localPage.deleteRecord(position));

    int newRecordVersion = 0;

    Assert.assertEquals(localPage.appendRecord(newRecordVersion,
        new byte[] {2, 2, 2, 4, 5, 6, 5, 4, 2, 2, 2}), position);

    int recordSize = localPage.getRecordSize(position);
    Assert.assertEquals(recordSize, 11);

    Assert.assertEquals(localPage.getRecordVersion(position), newRecordVersion);
    assertThat(localPage.getRecordBinaryValue(position, 0, recordSize))
        .isEqualTo(new byte[] {2, 2, 2, 4, 5, 6, 5, 4, 2, 2, 2});
  }

  @Test
  public void testDeleteAddBiggerVersion() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer,
        false);
    directCacheEntry.acquireExclusiveLock();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      deleteAddBiggerVersion(localPage);
      deleteAddBiggerVersion(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void deleteAddBiggerVersion(OClusterPage localPage) {
    int recordVersion = 0;
    recordVersion++;
    recordVersion++;

    int position =
        localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});

    Assert.assertTrue(localPage.deleteRecord(position));

    int newRecordVersion = 0;
    newRecordVersion++;
    newRecordVersion++;
    newRecordVersion++;
    newRecordVersion++;

    Assert.assertEquals(localPage.appendRecord(newRecordVersion,
        new byte[] {2, 2, 2, 4, 5, 6, 5, 4, 2, 2, 2}), position);

    int recordSize = localPage.getRecordSize(position);
    Assert.assertEquals(recordSize, 11);

    Assert.assertEquals(localPage.getRecordVersion(position), newRecordVersion);

    assertThat(localPage.getRecordBinaryValue(position, 0, recordSize))
        .isEqualTo(new byte[] {2, 2, 2, 4, 5, 6, 5, 4, 2, 2, 2});
  }

  @Test
  public void testDeleteAddEqualVersion() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();

    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      deleteAddEqualVersion(localPage);
      deleteAddEqualVersion(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void deleteAddEqualVersion(OClusterPage localPage) {
    int recordVersion = 0;
    recordVersion++;
    recordVersion++;

    int position = localPage.appendRecord(recordVersion,
        new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});

    Assert.assertTrue(localPage.deleteRecord(position));

    Assert.assertEquals(localPage.appendRecord(recordVersion,
        new byte[] {2, 2, 2, 4, 5, 6, 5, 4, 2, 2, 2}), position);

    int recordSize = localPage.getRecordSize(position);
    Assert.assertEquals(recordSize, 11);

    Assert.assertEquals(localPage.getRecordVersion(position), recordVersion);
    assertThat(localPage.getRecordBinaryValue(position, 0, recordSize)).
        isEqualTo(new byte[] {2, 2, 2, 4, 5, 6, 5, 4, 2, 2, 2});
  }

  @Test
  public void testDeleteAddEqualVersionKeepTombstoneVersion() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      deleteAddEqualVersionKeepTombstoneVersion(localPage);
      deleteAddEqualVersionKeepTombstoneVersion(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void deleteAddEqualVersionKeepTombstoneVersion(OClusterPage localPage) {
    int recordVersion = 0;
    recordVersion++;
    recordVersion++;

    int position =
        localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});

    Assert.assertTrue(localPage.deleteRecord(position));

    Assert.assertEquals(
        localPage.appendRecord(recordVersion, new byte[] {2, 2, 2, 4, 5, 6, 5, 4, 2, 2, 2}),
        position);

    int recordSize = localPage.getRecordSize(position);
    Assert.assertEquals(recordSize, 11);

    Assert.assertEquals(localPage.getRecordVersion(position), recordVersion);
    //    Assert.assertEquals(localPage.getRecordBinaryValue(position, 0, recordSize), new byte[] { 2, 2, 2, 4, 5, 6, 5, 4, 2, 2, 2 });
    assertThat(localPage.getRecordBinaryValue(position, 0, recordSize)).isEqualTo(new byte[] { 2, 2, 2, 4, 5, 6, 5, 4, 2, 2, 2 });
  }

  @Test
  public void testDeleteTwoOutOfFour() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer,
        false);
    directCacheEntry.acquireExclusiveLock();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      deleteTwoOutOfFour(localPage);
      deleteTwoOutOfFour(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void deleteTwoOutOfFour(OClusterPage localPage) {
    int recordVersion = 0;
    recordVersion++;

    int positionOne =
        localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});
    int positionTwo =
        localPage.appendRecord(recordVersion, new byte[] {2, 2, 3, 4, 5, 6, 5, 4, 3, 2, 2});
    int positionThree =
        localPage.appendRecord(recordVersion, new byte[] {3, 2, 3, 4, 5, 6, 5, 4, 3, 2, 3});
    int positionFour =
        localPage.appendRecord(recordVersion, new byte[] {4, 2, 3, 4, 5, 6, 5, 4, 3, 2, 4});

    Assert.assertEquals(localPage.getRecordsCount(), 4);
    Assert.assertEquals(positionOne, 0);
    Assert.assertEquals(positionTwo, 1);
    Assert.assertEquals(positionThree, 2);
    Assert.assertEquals(positionFour, 3);

    Assert.assertFalse(localPage.isDeleted(0));
    Assert.assertFalse(localPage.isDeleted(1));
    Assert.assertFalse(localPage.isDeleted(2));
    Assert.assertFalse(localPage.isDeleted(3));

    int freeSpace = localPage.getFreeSpace();

    Assert.assertTrue(localPage.deleteRecord(0));
    Assert.assertTrue(localPage.deleteRecord(2));

    Assert.assertFalse(localPage.deleteRecord(0));
    Assert.assertFalse(localPage.deleteRecord(7));

    Assert.assertEquals(localPage.findFirstDeletedRecord(0), 0);
    Assert.assertEquals(localPage.findFirstDeletedRecord(1), 2);
    Assert.assertEquals(localPage.findFirstDeletedRecord(3), -1);

    Assert.assertTrue(localPage.isDeleted(0));
    Assert.assertEquals(localPage.getRecordSize(0), -1);
    Assert.assertEquals(localPage.getRecordVersion(0), recordVersion);

    //    Assert.assertEquals(localPage.getRecordBinaryValue(1, 0, 11), new byte[] { 2, 2, 3, 4, 5, 6, 5, 4, 3, 2, 2 });

    assertThat(localPage.getRecordBinaryValue(1, 0, 11)).isEqualTo(new byte[] { 2, 2, 3, 4, 5, 6, 5, 4, 3, 2, 2 });
    Assert.assertEquals(localPage.getRecordSize(1), 11);
    Assert.assertEquals(localPage.getRecordVersion(1), recordVersion);

    Assert.assertTrue(localPage.isDeleted(2));
    Assert.assertEquals(localPage.getRecordSize(2), -1);
    Assert.assertEquals(localPage.getRecordVersion(2), recordVersion);

    //    Assert.assertEquals(localPage.getRecordBinaryValue(3, 0, 11), new byte[] { 4, 2, 3, 4, 5, 6, 5, 4, 3, 2, 4 });
    assertThat(localPage.getRecordBinaryValue(3, 0, 11)).isEqualTo(new byte[] { 4, 2, 3, 4, 5, 6, 5, 4, 3, 2, 4 });

    Assert.assertEquals(localPage.getRecordSize(3), 11);
    Assert.assertEquals(localPage.getRecordVersion(3), recordVersion);

    Assert.assertEquals(localPage.getRecordsCount(), 2);
    Assert.assertEquals(localPage.getFreeSpace(), freeSpace + 23 * 2);
  }

  @Test
  public void testAddFullPageDeleteAndAddAgain() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();

    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      addFullPageDeleteAndAddAgain(localPage);
      addFullPageDeleteAndAddAgain(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void addFullPageDeleteAndAddAgain(OClusterPage localPage) {
    Map<Integer, Byte> positionCounter = new HashMap<>();
    Set<Integer> deletedPositions = new HashSet<>();

    int lastPosition;
    byte counter = 0;
    int freeSpace = localPage.getFreeSpace();
    int recordVersion = 0;
    recordVersion++;

    do {
      lastPosition = localPage.appendRecord(recordVersion, new byte[] {counter, counter, counter});
      if (lastPosition >= 0) {
        Assert.assertEquals(lastPosition, positionCounter.size());
        positionCounter.put(lastPosition, counter);
        counter++;

        Assert.assertEquals(localPage.getFreeSpace(), freeSpace - (19 + ORecordVersionHelper.SERIALIZED_SIZE));
        freeSpace = localPage.getFreeSpace();
      }
    } while (lastPosition >= 0);

    int filledRecordsCount = positionCounter.size();
    Assert.assertEquals(localPage.getRecordsCount(), filledRecordsCount);

    for (int i = 0; i < filledRecordsCount; i += 2) {
      localPage.deleteRecord(i);
      deletedPositions.add(i);
      positionCounter.remove(i);
    }

    freeSpace = localPage.getFreeSpace();
    do {
      lastPosition = localPage.appendRecord(recordVersion, new byte[] { counter, counter, counter });
      if (lastPosition >= 0) {
        positionCounter.put(lastPosition, counter);
        counter++;

        Assert.assertEquals(localPage.getFreeSpace(), freeSpace - 15);
        freeSpace = localPage.getFreeSpace();
      }
    } while (lastPosition >= 0);

    Assert.assertEquals(localPage.getRecordsCount(), filledRecordsCount);
    for (Map.Entry<Integer, Byte> entry : positionCounter.entrySet()) {
      //      Assert.assertEquals(localPage.getRecordBinaryValue(entry.getKey(), 0, 3),
      //          new byte[] { entry.getValue(), entry.getValue(), entry.getValue() });

      assertThat(localPage.getRecordBinaryValue(entry.getKey(), 0, 3))
          .isEqualTo(new byte[] { entry.getValue(), entry.getValue(), entry.getValue() });

      Assert.assertEquals(localPage.getRecordSize(entry.getKey()), 3);

      if (deletedPositions.contains(entry.getKey()))
        Assert.assertEquals(localPage.getRecordVersion(entry.getKey()), recordVersion);

    }
  }

  @Test
  public void testAddBigRecordDeleteAndAddSmallRecords() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();

    try {
      final long seed = System.currentTimeMillis();

      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      addBigRecordDeleteAndAddSmallRecords(seed, localPage);
      addBigRecordDeleteAndAddSmallRecords(seed, directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void addBigRecordDeleteAndAddSmallRecords(long seed, OClusterPage localPage) {
    final Random mersenneTwisterFast = new Random(seed);

    int recordVersion = 0;
    recordVersion++;
    recordVersion++;

    final byte[] bigChunk = new byte[OClusterPage.MAX_ENTRY_SIZE / 2];

    mersenneTwisterFast.nextBytes(bigChunk);

    int position = localPage.appendRecord(recordVersion, bigChunk);
    Assert.assertEquals(position, 0);
    Assert.assertEquals(localPage.getRecordVersion(0), recordVersion);

    Assert.assertTrue(localPage.deleteRecord(0));

    recordVersion++;
    int freeSpace = localPage.getFreeSpace();
    Map<Integer, Byte> positionCounter = new HashMap<>();
    int lastPosition;
    byte counter = 0;
    do {
      lastPosition = localPage.appendRecord(recordVersion, new byte[] { counter, counter, counter });
      if (lastPosition >= 0) {
        Assert.assertEquals(lastPosition, positionCounter.size());
        positionCounter.put(lastPosition, counter);
        counter++;

        if (lastPosition == 0)
          Assert.assertEquals(localPage.getFreeSpace(), freeSpace - 15);
        else
          Assert.assertEquals(localPage.getFreeSpace(), freeSpace - (19 + ORecordVersionHelper.SERIALIZED_SIZE));

        freeSpace = localPage.getFreeSpace();
      }
    } while (lastPosition >= 0);

    Assert.assertEquals(localPage.getRecordsCount(), positionCounter.size());
    for (Map.Entry<Integer, Byte> entry : positionCounter.entrySet()) {
      //      Assert.assertEquals(localPage.getRecordBinaryValue(entry.getKey(), 0, 3),
      //          new byte[] { entry.getValue(), entry.getValue(), entry.getValue() });

      assertThat(localPage.getRecordBinaryValue(entry.getKey(), 0, 3))
          .isEqualTo(new byte[] { entry.getValue(), entry.getValue(), entry.getValue() });
      Assert.assertEquals(localPage.getRecordSize(entry.getKey()), 3);
      Assert.assertEquals(localPage.getRecordVersion(entry.getKey()), recordVersion);
    }
  }

  @Test
  public void testFindFirstRecord() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();

    final long seed = System.currentTimeMillis();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      findFirstRecord(seed, localPage);
      findFirstRecord(seed, directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void findFirstRecord(long seed, OClusterPage localPage) {
    final Random mersenneTwister = new Random(seed);
    Set<Integer> positions = new HashSet<>();

    int lastPosition;
    byte counter = 0;
    int freeSpace = localPage.getFreeSpace();

    int recordVersion = 0;
    recordVersion++;

    do {
      lastPosition = localPage.appendRecord(recordVersion, new byte[] { counter, counter, counter });
      if (lastPosition >= 0) {
        Assert.assertEquals(lastPosition, positions.size());
        positions.add(lastPosition);
        counter++;

        Assert.assertEquals(localPage.getFreeSpace(), freeSpace - (19 + ORecordVersionHelper.SERIALIZED_SIZE));
        freeSpace = localPage.getFreeSpace();
      }
    } while (lastPosition >= 0);

    int filledRecordsCount = positions.size();
    Assert.assertEquals(localPage.getRecordsCount(), filledRecordsCount);

    for (int i = 0; i < filledRecordsCount; i++) {
      if (mersenneTwister.nextBoolean()) {
        localPage.deleteRecord(i);
        positions.remove(i);
      }
    }

    int recordsIterated = 0;
    int recordPosition = 0;
    int lastRecordPosition = -1;

    do {
      recordPosition = localPage.findFirstRecord(recordPosition);
      if (recordPosition < 0)
        break;

      Assert.assertTrue(positions.contains(recordPosition));
      Assert.assertTrue(recordPosition > lastRecordPosition);

      lastRecordPosition = recordPosition;

      recordPosition++;
      recordsIterated++;
    } while (recordPosition >= 0);

    Assert.assertEquals(recordsIterated, positions.size());
  }

  @Test
  public void testFindLastRecord() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();

    final long seed = System.currentTimeMillis();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      findLastRecord(seed, localPage);
      findLastRecord(seed, directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void findLastRecord(long seed, OClusterPage localPage) {
    final Random mersenneTwister = new Random(seed);
    Set<Integer> positions = new HashSet<>();

    int lastPosition;
    byte counter = 0;
    int freeSpace = localPage.getFreeSpace();

    int recordVersion = 0;
    recordVersion++;

    do {
      lastPosition = localPage.appendRecord(recordVersion, new byte[] { counter, counter, counter });
      if (lastPosition >= 0) {
        Assert.assertEquals(lastPosition, positions.size());
        positions.add(lastPosition);
        counter++;

        Assert.assertEquals(localPage.getFreeSpace(), freeSpace - (19 + ORecordVersionHelper.SERIALIZED_SIZE));
        freeSpace = localPage.getFreeSpace();
      }
    } while (lastPosition >= 0);

    int filledRecordsCount = positions.size();
    Assert.assertEquals(localPage.getRecordsCount(), filledRecordsCount);

    for (int i = 0; i < filledRecordsCount; i++) {
      if (mersenneTwister.nextBoolean()) {
        localPage.deleteRecord(i);
        positions.remove(i);
      }
    }

    int recordsIterated = 0;
    int recordPosition = Integer.MAX_VALUE;
    int lastRecordPosition = Integer.MAX_VALUE;
    do {
      recordPosition = localPage.findLastRecord(recordPosition);
      if (recordPosition < 0)
        break;

      Assert.assertTrue(positions.contains(recordPosition));
      Assert.assertTrue(recordPosition < lastRecordPosition);

      recordPosition--;
      recordsIterated++;
    } while (recordPosition >= 0);

    Assert.assertEquals(recordsIterated, positions.size());
  }

  @Test
  public void testSetGetNextPage() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      setGetNextPage(localPage);
      setGetNextPage(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void setGetNextPage(OClusterPage localPage) {
    localPage.setNextPage(1034);
    Assert.assertEquals(localPage.getNextPage(), 1034);
  }

  @Test
  public void testSetGetPrevPage() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      setGetPrevPage(localPage);
      setGetPrevPage(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void setGetPrevPage(OClusterPage localPage) {
    localPage.setPrevPage(1034);
    Assert.assertEquals(localPage.getPrevPage(), 1034);
  }

  @Test
  public void testReplaceOneRecordWithBiggerSize() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();

    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      replaceOneRecordWithBiggerSize(localPage);
      replaceOneRecordWithBiggerSize(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void replaceOneRecordWithBiggerSize(OClusterPage localPage) {
    Assert.assertEquals(localPage.getRecordsCount(), 0);

    int recordVersion = 0;
    recordVersion++;

    int index = localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});
    int freeSpace = localPage.getFreeSpace();

    int newRecordVersion = recordVersion;
    newRecordVersion++;

    int written = localPage.replaceRecord(index, new byte[] {5, 2, 3, 4, 5, 11, 5, 4, 3, 2, 1, 3},
        newRecordVersion);
    Assert.assertEquals(localPage.getFreeSpace(), freeSpace);
    Assert.assertEquals(written, 11);

    Assert.assertEquals(localPage.getRecordSize(index), 11);
    assertThat(localPage.getRecordBinaryValue(index, 0, 11))
        .isEqualTo(new byte[] {5, 2, 3, 4, 5, 11, 5, 4, 3, 2, 1});
    Assert.assertEquals(localPage.getRecordVersion(index), newRecordVersion);
  }

  @Test
  public void testReplaceOneRecordWithEqualSize() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      replaceOneRecordWithEqualSize(localPage);
      replaceOneRecordWithEqualSize(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void replaceOneRecordWithEqualSize(OClusterPage localPage) {
    Assert.assertEquals(localPage.getRecordsCount(), 0);

    int recordVersion = 0;
    recordVersion++;

    int index = localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});
    int freeSpace = localPage.getFreeSpace();

    int newRecordVersion = recordVersion;
    newRecordVersion++;

    int written = localPage.replaceRecord(index, new byte[] {5, 2, 3, 4, 5, 11, 5, 4, 3, 2, 1},
        newRecordVersion);
    Assert.assertEquals(localPage.getFreeSpace(), freeSpace);
    Assert.assertEquals(written, 11);

    Assert.assertEquals(localPage.getRecordSize(index), 11);

    assertThat(localPage.getRecordBinaryValue(index, 0, 11))
        .isEqualTo(new byte[] {5, 2, 3, 4, 5, 11, 5, 4, 3, 2, 1});
    Assert.assertEquals(localPage.getRecordVersion(index), newRecordVersion);
  }

  @Test
  public void testReplaceOneRecordWithSmallerSize() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer,
        false);
    directCacheEntry.acquireExclusiveLock();

    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      replaceOneRecordWithSmallerSize(localPage);
      replaceOneRecordWithSmallerSize(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void replaceOneRecordWithSmallerSize(OClusterPage localPage) {
    Assert.assertEquals(localPage.getRecordsCount(), 0);

    int recordVersion = 0;
    recordVersion++;

    int index = localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});
    int freeSpace = localPage.getFreeSpace();

    int newRecordVersion = recordVersion;
    newRecordVersion++;

    int written = localPage.replaceRecord(index, new byte[] {5, 2, 3, 4, 5, 11,}, newRecordVersion);
    Assert.assertEquals(localPage.getFreeSpace(), freeSpace);
    Assert.assertEquals(written, 6);

    Assert.assertEquals(localPage.getRecordSize(index), 6);

    assertThat(localPage.getRecordBinaryValue(index, 0, 6)).
        isEqualTo(new byte[] {5, 2, 3, 4, 5, 11});
    Assert.assertEquals(localPage.getRecordVersion(index), newRecordVersion);
  }

  @Test
  public void testReplaceOneRecordNoVersionUpdate() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer,
        false);
    directCacheEntry.acquireExclusiveLock();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      replaceOneRecordNoVersionUpdate(localPage);
      replaceOneRecordNoVersionUpdate(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void replaceOneRecordNoVersionUpdate(OClusterPage localPage) {
    Assert.assertEquals(localPage.getRecordsCount(), 0);

    int recordVersion = 0;
    recordVersion++;

    int index = localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});
    int freeSpace = localPage.getFreeSpace();

    int newRecordVersion = recordVersion + 1;

    int written = localPage.replaceRecord(index, new byte[] {5, 2, 3, 4, 5, 11, 5, 4, 3, 2, 1, 3},
        -1);
    Assert.assertEquals(localPage.getFreeSpace(), freeSpace);
    Assert.assertEquals(written, 11);

    Assert.assertEquals(localPage.getRecordSize(index), 11);

    assertThat(localPage.getRecordBinaryValue(index, 0, 11)).
        isEqualTo(new byte[] {5, 2, 3, 4, 5, 11, 5, 4, 3, 2, 1});
    Assert.assertEquals(localPage.getRecordVersion(index), recordVersion);
  }

  @Test
  public void testReplaceOneRecordLowerVersion() {
    OByteBufferPool bufferPool = OByteBufferPool.instance(null);
    OPointer pointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(pointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();

    OPointer directPointer = bufferPool.acquireDirect(true);
    OCachePointer directCachePointer = new OCachePointer(directPointer, bufferPool, 0, 0);
    directCachePointer.incrementReferrer();

    OCacheEntry directCacheEntry = new OCacheEntryImpl(0, 0, directCachePointer, false);
    directCacheEntry.acquireExclusiveLock();
    try {
      OClusterPage localPage = new OClusterPage(new OCacheEntryChanges(cacheEntry), true);
      OClusterPage directLocalPage = new OClusterPage(directCacheEntry, true);

      replaceOneRecordLowerVersion(localPage);
      replaceOneRecordLowerVersion(directLocalPage);

      assertChangesTracking(localPage, directPointer, bufferPool);
    } finally {
      cacheEntry.releaseExclusiveLock();
      directCacheEntry.releaseExclusiveLock();

      cachePointer.decrementReferrer();
      directCachePointer.decrementReferrer();
    }
  }

  private static void replaceOneRecordLowerVersion(OClusterPage localPage) {
    Assert.assertEquals(localPage.getRecordsCount(), 0);

    int recordVersion = 0;
    recordVersion++;

    int index = localPage.appendRecord(recordVersion, new byte[] {1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1});
    int freeSpace = localPage.getFreeSpace();

    int newRecordVersion = recordVersion;

    int written = localPage.replaceRecord(index, new byte[] {5, 2, 3, 4, 5, 11, 5, 4, 3, 2, 1, 3},
        newRecordVersion);
    Assert.assertEquals(localPage.getFreeSpace(), freeSpace);
    Assert.assertEquals(written, 11);

    Assert.assertEquals(localPage.getRecordSize(index), 11);

    assertThat(localPage.getRecordBinaryValue(index, 0, 11)).
        isEqualTo(new byte[] {5, 2, 3, 4, 5, 11, 5, 4, 3, 2, 1});
    Assert.assertEquals(localPage.getRecordVersion(index), recordVersion);
  }

  private static void assertChangesTracking(OClusterPage localPage, OPointer pointer,
      OByteBufferPool bufferPool) {
    OPointer restoredPointer = bufferPool.acquireDirect(true);

    OCachePointer cachePointer = new OCachePointer(restoredPointer, bufferPool, 0, 0);
    cachePointer.incrementReferrer();

    OCacheEntry cacheEntry = new OCacheEntryImpl(0, 0, cachePointer, false);
    cacheEntry.acquireExclusiveLock();
    try {
      OClusterPage restoredPage = new OClusterPage(cacheEntry, false);

      OWALChanges changes = localPage.getChanges();
      restoredPage.restoreChanges(changes);

      assertThat(getBytes(restoredPointer.getNativeByteBuffer(),
          OClusterPage.PAGE_SIZE - SYSTEM_OFFSET))
          .isEqualTo(getBytes(pointer.getNativeByteBuffer(),
              OClusterPage.PAGE_SIZE - SYSTEM_OFFSET));
    } finally {
      cacheEntry.releaseExclusiveLock();
      cachePointer.decrementReferrer();
    }
  }

  private static byte[] getBytes(ByteBuffer buffer, int len) {
    byte[] result = new byte[len];
    buffer.position(ClusterPageTest.SYSTEM_OFFSET);
    buffer.get(result);

    return result;
  }
}
